from pgzero import clock

import numpy as np

from pw26flow.maps import (
    get_layer,
    get_goal_pos,
    fixed_nodes,
    random_starting_point,
    FLOW_DIR,
    ZONE_NONE,
    ZONE_TARGET,
    ANIMALS,
)


MAX_NODES = 100
MAX_EDGES = 100
MAX_AGENTS = 10
MAX_PARTICLES = 150

TILE_SIZE = 64

FLOW_STRENGTH = 10
AGENT_SPEED = 25

CONNECT_DISTANCE = 20


class InventoryItem:
    def __init__(self, item_type, **kwargs):
        self.item_type = item_type
        for key, value in kwargs.items():
            setattr(self, key, value)


class GameState:
    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.finished = False
        self.time = 0
        self.map_data = None
        self.start_node = 0
        self.next_node = 0
        self.next_edge = 0
        self.next_agent = 0
        self.next_particle = 0
        self.node_pos = np.zeros((MAX_NODES, 2))
        self.node_vel = np.zeros((MAX_NODES, 2))
        self.node_mass = np.zeros(MAX_NODES)
        self.edges = np.zeros((MAX_EDGES, 2), dtype=np.uint32)
        self.lengths = np.zeros(MAX_EDGES)
        self.stiffness = np.zeros(MAX_EDGES)
        self.agent_pos = np.zeros((MAX_AGENTS, 2))
        self.agent_vel = np.zeros((MAX_AGENTS, 2))
        self.agent_state = np.zeros(MAX_AGENTS, dtype=np.uint32)
        self.agent_edge = np.zeros(MAX_AGENTS, dtype=np.uint32)
        self.agent_tex = np.zeros(MAX_AGENTS, dtype=np.uint8)
        self.particle_pos = np.zeros((MAX_PARTICLES, 2))
        self.cursor_pos = np.array([0, 0])
        self.active_inventory = None
        self.flow = None
        self.zones = None
        self.target_marker = None

    def add_node(self, pos, mass=0.0):
        if mass == 0.0 and not self.get_flow_at(pos).any():
            return
        node_id = self.next_node
        self.node_pos[node_id] = pos
        self.node_mass[node_id] = mass
        self.next_node = node_id + 1
        return node_id

    def connect_nodes(self, from_node, to_node, mass=1.0, stiffness=1.0):
        """ Connect two nodes, mass is distributed to both nodes. """
        edge_id = self.next_edge
        self.edges[edge_id] = (from_node, to_node)
        self.lengths[edge_id] = np.linalg.norm(
            self.node_pos[from_node] - self.node_pos[to_node]
        )
        half_mass = mass / 2
        self.node_mass[[from_node, to_node]] += half_mass
        self.stiffness[edge_id] = stiffness
        self.next_edge = edge_id + 1
        if edge_id == 0:
            self.notify_agent()
        return edge_id

    def find_closest_node(self, pos, max_distance=None):
        distances = np.abs(
            np.linalg.norm(self.node_pos[: self.next_node] - pos, axis=1)
        )
        closest = distances.argmin()
        if max_distance is None or distances[closest] <= max_distance:
            return closest

    def add_agent(self, pos):
        agent_id = self.next_agent
        self.agent_pos[agent_id] = pos
        self.agent_vel[agent_id] = (
            np.random.rand(2).clip(0.5) * np.random.choice([-1, 1], 2) * AGENT_SPEED
        )
        self.agent_tex[agent_id] = np.random.randint(len(ANIMALS))
        self.next_agent = agent_id + 1
        return agent_id

    def update_positions(self, dt):
        if not self.finished:
            self.time += dt
        # Update flow field and set velocities
        strength = FLOW_STRENGTH + (FLOW_STRENGTH / 2 * np.sin(self.time))
        for i in range(self.next_node):
            self.node_vel[i] = self.get_flow_at(self.node_pos[i]) * strength
        for i in range(self.next_particle):
            vel = self.get_flow_at(self.particle_pos[i]) * strength * dt
            self.particle_pos[i] += vel
            pos = self.particle_pos[i]
            if (
                not vel.any()
                or pos[0] < 0
                or pos[0] > self.width
                or pos[1] < 0
                or pos[1] > self.height
            ):
                self.particle_pos[i] = random_starting_point(
                    self.map_data, area_name="particle-start"
                )
        if self.next_edge == 0:
            return
        # Update forces from springs to set velocity
        cur_edges = self.edges[: self.next_edge]
        cur_edges_pos = self.node_pos[cur_edges]
        cur_edges_dir = cur_edges_pos[:, 1] - cur_edges_pos[:, 0]
        cur_lengths = np.linalg.norm(cur_edges_dir, axis=1)
        length_changes = cur_lengths - self.lengths[: self.next_edge]
        edge_diff = (
            cur_edges_dir / cur_lengths[:, np.newaxis] * length_changes[:, np.newaxis]
        )
        forces = self.stiffness[: self.next_edge, np.newaxis] * edge_diff
        # Apply forces sequentially, otherwise nodes receiving multiple forces
        # in a single computation cause problems (only one force is applied then)
        for i, force in enumerate(forces):
            edge = cur_edges[i]
            self.node_vel[edge[0]] += force / self.node_mass[edge[0]]
            self.node_vel[edge[1]] -= force / self.node_mass[edge[1]]
        # Update positions based on velocity
        self.node_pos[: self.next_node] += self.node_vel[: self.next_node] * dt

    def update_agent_positions(self, dt):
        freely_moving = self.agent_state == 0
        head_to_start = self.agent_state == 1
        moving_to_start = self.agent_state == 2
        move_along_edge = self.agent_state == 3
        in_target_area = self.agent_state == 4

        if not self.finished and in_target_area.all():
            self.finished = True

        start_pos = self.node_pos[self.start_node]
        if head_to_start.any():
            directions = start_pos - self.agent_pos[head_to_start]
            self.agent_vel[head_to_start] = (
                directions
                / np.linalg.norm(directions, axis=1)[np.newaxis, :].T
                * AGENT_SPEED
                * (np.random.rand() + 0.5)
            )
            self.agent_state[head_to_start] = 2
        if moving_to_start.any():
            distances = np.linalg.norm(
                start_pos - self.agent_pos[moving_to_start], axis=1
            )
            at_start = distances <= 5
            if at_start.any():
                indices = np.arange(MAX_AGENTS)[moving_to_start][at_start]
                self.agent_vel[indices] = [self.start_node, 0.0]
                self.agent_edge[indices] = [
                    self.get_edge_from_node(self.start_node) for _ in indices
                ]
                self.agent_state[indices] = 3
        if move_along_edge.any():
            indices = np.arange(MAX_AGENTS)[move_along_edge]
            self.agent_vel[indices, 1] += 0.5 * dt
            for i in indices:
                cur_edge = self.agent_edge[i]
                from_node = int(self.agent_vel[i, 0])
                to_node = self.get_target_node(cur_edge, from_node)
                if self.agent_vel[i, 1] > 1.0:
                    if self.get_zone_at(self.agent_pos[i]) == ZONE_TARGET:
                        self.agent_state[i] = 4
                        self.agent_vel[i] = (
                            np.random.rand(2).clip(0.5)
                            * np.random.choice([-1, 1], 2)
                            * AGENT_SPEED
                        )
                        continue
                    next_edge = self.get_edge_from_node(to_node, avoid=[cur_edge])
                    self.agent_edge[i] = next_edge
                    self.agent_vel[i] = [to_node, 0.0]
                    from_node = to_node
                    to_node = self.get_target_node(next_edge, to_node)
                from_pos = self.node_pos[from_node]
                to_pos = self.node_pos[to_node]
                self.agent_pos[i] = (
                    from_pos + (to_pos - from_pos) * self.agent_vel[i, 1]
                )

        not_on_edge = np.logical_not(move_along_edge)
        self.agent_pos[not_on_edge] += self.agent_vel[not_on_edge] * dt

        for i, pos in enumerate(self.agent_pos):
            if self.agent_state[i] not in [0, 4]:
                continue
            if (
                pos[0] < 0
                or pos[0] > self.width
                or pos[1] < 0
                or pos[1] > self.height
                or self.get_flow_at(pos).any()
            ):
                self.agent_pos[i] -= self.agent_vel[i] * dt
                self.agent_vel[i] = (
                    np.random.rand(2).clip(0.5)
                    * np.random.choice([-1, 1], 2)
                    * AGENT_SPEED
                )

    def rotate_active_inventory(self, degrees):
        if self.active_inventory is not None:
            item = self.active_inventory
            theta = np.radians(degrees)
            c, s = np.cos(theta), np.sin(theta)
            rotation = np.array(((c, s), (-s, c)))
            if item.item_type == "connection":
                item.line = item.line @ rotation

    def load_map(self, map_data):
        self.map_data = map_data
        self.flow = get_layer(map_data, "flow")
        self.zones = get_layer(map_data, "zones")
        self.tex1 = get_layer(map_data, "tex1")
        for pos, is_start in fixed_nodes(map_data):
            i = self.add_node(pos, mass=1e6)
            if is_start:
                self.start_node = i
        self.target_marker = get_goal_pos(map_data)
        for i in range(MAX_AGENTS):
            self.add_agent(random_starting_point(map_data))
        clock.schedule(self.add_particle, 0.5)

    def get_flow_at(self, pos):
        if self.flow is None:
            return np.zeros(2)
        x = int(pos[0] / TILE_SIZE)
        y = int(pos[1] / TILE_SIZE)
        if x < 0 or x >= self.flow.shape[1] or y < 0 or y >= self.flow.shape[0]:
            return np.zeros(2)
        # TODO - bilinear interpolation?
        return FLOW_DIR[self.flow[y, x]]

    def get_zone_at(self, pos):
        if self.zones is None:
            return ZONE_NONE
        x = int(pos[0] / TILE_SIZE)
        y = int(pos[1] / TILE_SIZE)
        if x < 0 or x >= self.zones.shape[1] or y < 0 or y >= self.zones.shape[0]:
            return ZONE_NONE
        return self.zones[y, x]

    def get_edge_from_node(self, node_id, avoid=None):
        edges = set()
        for i, edge in enumerate(self.edges):
            if i == self.next_edge:
                break
            if edge[0] == node_id or edge[1] == node_id:
                edges.add(i)
        if len(edges) > 1 and avoid is not None:
            edges = edges - set(avoid)
        return np.random.choice(list(edges))

    def get_target_node(self, current_edge, from_node):
        edge = self.edges[current_edge]
        if edge[0] == from_node:
            return edge[1]
        else:
            return edge[0]

    def notify_agent(self):
        for i in range(self.next_agent):
            if self.agent_state[i] < 1:
                self.agent_state[i] = 1
                clock.schedule(self.notify_agent, 1.0)
                return

    def add_particle(self):
        if self.next_particle == MAX_PARTICLES:
            return
        self.particle_pos[self.next_particle] = random_starting_point(
            self.map_data, area_name="particle-start"
        )
        self.next_particle += 1
        clock.schedule(self.add_particle, 0.33)
